@@ -1,6 +1,7 @@
 import io
 import json
 import os
+import re
 from datetime import datetime, timedelta
 
 import google_auth_oauthlib.flow
@@ -24,6 +25,7 @@ from plaid.model.institutions_get_by_id_request import InstitutionsGetByIdReques
 from plaid.model.institutions_get_request import InstitutionsGetRequest
 from plaid.model.item_get_request import ItemGetRequest
 from plaid.model.item_public_token_exchange_request import ItemPublicTokenExchangeRequest
+from plaid.model.item_remove_request import ItemRemoveRequest
 from plaid.model.link_token_create_request import LinkTokenCreateRequest
 from plaid.model.link_token_create_request_user import LinkTokenCreateRequestUser
 from plaid.model.products import Products
@@ -39,19 +41,31 @@ else:
     session_manager = FlaskSessionManager(session)
     print('Using Flask session manager')
 
+@app.before_request
+def load_session():
+    if request.endpoint in ('login', 'sign_in_with_google_callback'):
+        return
+    if not session_manager.user_id:
+        user_id = request.cookies.get('user_id')
+        if not user_id:
+            return redirect(url_for('login'))
+        session_manager.register_user_id(user_id)
+
 @app.before_first_request
 def initialize_app():
-    env_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), '.env')
-    if os.path.isfile(env_file):
-        load_dotenv()
-    elif os.environ.get('GOOGLE_CLOUD_PROJECT'):
+    required_env_variables = ('GOOGLE_CLOUD_CLIENT_ID', 'GOOGLE_CLOUD_CLIENT_CONFIG', 'FLASK_SECRET_KEY')
+    if os.environ.get('GOOGLE_CLOUD_PROJECT'):
         project_id = os.environ.get('GOOGLE_CLOUD_PROJECT')
         secrets_client = secretmanager.SecretManagerServiceClient()
         name = f'projects/{project_id}/secrets/app_engine_python_env/versions/latest'
         payload = secrets_client.access_secret_version(name=name).payload.data.decode('UTF-8')
         load_dotenv(stream=io.StringIO(payload))
     else:
-        raise Exception("No local .env or GOOGLE_CLOUD_PROJECT detected. No secrets found.")
+        # When running a web server not hosted in Google Cloud (eg. locally), we
+        # must authenticate with Google Cloud APIs.
+        required_env_variables += ('GOOGLE_APPLICATION_CREDENTIALS',)
+    if not all(env in os.environ for env in required_env_variables):
+        raise EnvironmentError(f"The following environment variables are required: {required_env_variables}")
 
 @app.route('/login')
 def login():
@@ -60,11 +74,6 @@ def login():
 @app.route('/')
 @app.route('/index')
 def index():
-    if not session_manager.user_id:
-        user_id = request.cookies.get('user_id')
-        if not user_id:
-            return redirect(url_for('login'))
-        session_manager.register_user_id(user_id)
     try:
         session_data = session_manager.get_session()
     except Exception:
@@ -135,7 +144,7 @@ def unquote(url):
 @app.route('/authorize-google-credentials')
 def authorize_google_credentials():
     # Create flow instance to manage the OAuth 2.0 Authorization Grant Flow steps.
-    client_config = json.loads(os.environ.get('GOOGLE_CLOUD_CLIENT_CONFIG', {}))
+    client_config = parse_google_cloud_client_config()
     flow = google_auth_oauthlib.flow.Flow.from_client_config(client_config, GOOGLE_SCOPES)
 
     # The URI created here must exactly match one of the authorized redirect URIs
@@ -161,7 +170,7 @@ def google_oauth_callback():
     # Specify the state when creating the flow in the callback so that it can
     # verified in the authorization server response.
     state = request.cookies.get('google_oauth_state')
-    client_config = json.loads(os.environ.get('GOOGLE_CLOUD_CLIENT_CONFIG', {}))
+    client_config = parse_google_cloud_client_config()
     flow = google_auth_oauthlib.flow.Flow.from_client_config(client_config, GOOGLE_SCOPES, state=state)
     flow.redirect_uri = url_for('google_oauth_callback', _external=True)
 
@@ -217,10 +226,11 @@ def manage_plaid_items():
         <br><br>
         <a href={url_for('index')}>Back to home</a>
         """
-    access_tokens = get_plaid_items(session_data).values()
+    access_tokens = get_plaid_items(session_data, remove_inactive_items=False).values()
     item_info = get_plaid_item_info(access_tokens, session_data)
+    plaid_env = session_data.get('plaid_env', 'sandbox')
     resp = make_response(render_template('plaid_items_form.html', plaid_link_token=link_token,
-        plaid_oauth_redirect=False, plaid_items=item_info))
+        plaid_oauth_redirect=False, plaid_items=item_info, plaid_env=plaid_env))
     resp.set_cookie('plaid_link_token', link_token, httponly=True)
     return resp
 
@@ -268,6 +278,19 @@ def sync():
     session_manager['last_sync'] = datetime.now().strftime(TIMESTAMP_FORMAT)
     return redirect(url_for('index'))
 
+@app.route('/remove-plaid-item')
+def remove_plaid_item():
+    token = request.args.get('access_token')
+    session_data = session_manager.get_session()
+    plaid_client = build_plaid_client(session_data)
+    item_id = (plaid_client.item_get(ItemGetRequest(token)))['item']['item_id']
+    plaid_request = ItemRemoveRequest(access_token=token)
+    plaid_client.item_remove(plaid_request)
+    del session_data['plaid_items'][item_id]
+    session_manager.set_session(session_data)
+    redirect_url = request.args.get('redirect_url', url_for('index'))
+    return redirect(redirect_url)
+
 def status_check(session_data: dict) -> dict:
     plaid_env = session_data.get('plaid_env', 'sandbox')
     plaid_items = get_plaid_items(session_data)
@@ -279,6 +302,19 @@ def status_check(session_data: dict) -> dict:
         'spreadsheet_url': session_data.get(f'{plaid_env}_spreadsheet_url'),
     }
 
+def parse_google_cloud_client_config() -> dict:
+    env_variable = os.environ.get('GOOGLE_CLOUD_CLIENT_CONFIG')
+    if not env_variable:
+        raise EnvironmentError('Expected to find GOOGLE_CLOUD_CLIENT_CONFIG environment variable.')
+    if os.path.isfile(env_variable):
+        with open(env_variable, 'r') as file:
+            client_config = json.load(file)
+    else:
+        try:
+            client_config = json.loads(env_variable)
+        except json.JSONDecodeError:
+            raise ValueError('GOOGLE_CLOUD_CLIENT_CONFIG must be a valid filepath or valid JSON.')
+    return client_config
 
 def user_allowed_sync(session_data: dict) -> bool:
     last_sync = session_data.get('last_sync')
@@ -312,10 +348,12 @@ def validate_plaid_access_tokens(plaid_items: dict, session_data: dict) -> bool:
             return False
     return True
 
-def get_plaid_items(session_data: dict) -> dict:
+def get_plaid_items(session_data: dict, remove_inactive_items: bool = True) -> dict:
     plaid_env = session_data.get('plaid_env', 'sandbox')
     plaid_items = session_data.get('plaid_items', {})
-    return {k: v for k, v in plaid_items.items() if v.lower().startswith(f'access-{plaid_env}')}
+    if remove_inactive_items:
+        return {k: v for k, v in plaid_items.items() if v.lower().startswith(f'access-{plaid_env}')}
+    return plaid_items
 
 def lookup_spreadsheet_name_for_env(
         plaid_env: str,
@@ -418,12 +456,23 @@ def get_plaid_item_info(access_tokens: list, session_data: dict) -> tuple:
         try:
             response = plaid_client.item_get(ItemGetRequest(token))
             ins_id = response['item']['institution_id']
+            healthy_state = response['item']['error'] == None
             ins_request = InstitutionsGetByIdRequest(ins_id, [CountryCode('US')])
             response = plaid_client.institutions_get_by_id(ins_request)
+            ins_name = response['institution']['name']
             link_update_token = request_link_update_token(plaid_client, token, session_data)
-            results.append((response['institution']['name'], True, link_update_token))
+            token_env = re.findall(r"access-(\w+)-.*", token)[0]
         except PlaidApiException as e:
-            raise e
+            error_code = json.loads(e.body)['error_code']
+            if error_code == 'INVALID_ACCESS_TOKEN':
+                token_env_regex = re.findall(r"access-(\w+)-.*", token)
+                token_env = token_env_regex[0] if token_env_regex else 'unknown'
+                ins_name = f'{token_env} institution'
+                healthy_state = None
+                link_update_token = None
+            else:
+                raise e
+        results.append((ins_name, token_env, healthy_state, link_update_token, token))
     return results
 
 @app.route('/revoke-google-credentials')
@@ -449,4 +498,4 @@ def run_web_server(host: str = 'localhost', port: int = 8080, debug: bool = Fals
     app.run(host=host, port=port, debug=debug, load_dotenv=False, **kwargs)
 
 if __name__ == '__main__':
-    run_web_server()
+    run_web_server(debug=True, ssl_context='adhoc')
